/* File: autocomplete.cpp
 * Course: CS216-00x
 * Project: Project 2
 * Purpose: the implementation of member functions for the class named Autocomplete.
 *
 */
#include <iostream>
#include "autocomplete.h"

// default constructor
Autocomplete::Autocomplete()
{
    // You can leave it blank
    // since the private data member terms has been initialized through default constructor of SortingList class 
}


// inserts the newterm to the sequence
// note that a SortingList<Term> object, named terms
//          represents the sequence
void Autocomplete::insert(Term newterm)
{
	terms.insert(newterm);
}


// sort all the terms by query in the lexicographical order  
void Autocomplete::sort()
{
	terms.sort();
}

// binary search helper function
// as private function
// since it serves binary_search member function only
int Autocomplete::binary_searchHelper(SortingList<Term> terms, string key, int left_index, int right_index)
{
    if(right_index < left_index)
    {
        int KEY_NOT_FOUND = -1;
        return KEY_NOT_FOUND;
    }
    else
    {
        Term keyterm(key, 0);
        int r = key.length();
        int middle = (left_index + right_index) / 2;
        if(Term::compareByPrefix(keyterm,terms[middle],r) > 0)
        {
            return binary_searchHelper(terms, key, left_index, middle - 1);

        }
        else if(Term::compareByPrefix(keyterm,terms[middle],r)  < 0)
        {
            return binary_searchHelper(terms, key, middle + 1, right_index);
        }
        else
        {
            return middle;
        }
    }
}

// return the index number of the term whose query 
// prefix-matches the given prefix, using binary search algorithm
// Note that binary search can only be applied to sorted sequence
// Note that you may want a binary search helper function
int Autocomplete::binary_search(string prefix)
{
	Term keyterm(prefix, 0);
	terms.sort();
	
	int index = -1;
	int r = prefix.length();
	int firstIndex = 0;
	int lastIndex = terms.size() - 1;

	while (firstIndex <= lastIndex)
	{
		int mid = (firstIndex + lastIndex)/2;
		if (Term::compareByPrefix(terms[mid], keyterm, r) == 0)
		{
			return mid;
		}
		else if (Term::compareByPrefix(terms[mid], keyterm, r) == 1)
		{
			firstIndex = mid + 1;
		}
		else
		{
			lastIndex = mid -1;
		}
	}
	return index ;
}

// first: the index of the first query whose prefix matches
//          the search key, or -1 if no such key
// last: the index of the last query whose prefix matches 
//         the search key, or -1 if no such key
// key: the given prefix to match
// hints: you can call binary_search() first to find one matched index number,
//        say hit, then look up and down from hit, to find first and last respectively
void Autocomplete::search(string key, int& firstIndex, int& lastIndex)
{
	int index = this->binary_search(key);

	if (index == -1)
	{
		firstIndex = -1;
		lastIndex = -1;
		return;
	}
	
	int i = index;
	while ( i >= 0 && Term::compareByPrefix(terms[i], terms[index], key.size()) == 0)
	{
		i--;
	}
	firstIndex = i + 1;

	int k = index;
	while ( i < terms.size() && Term::compareByPrefix(terms[k], terms[index], key.size()) == 0)
	{
		k++;
	}
	lastIndex = k -1;
}

// returns all terms that start with the given prefix, in descending order of weight
SortingList<Term> Autocomplete::allMatches(string prefix)
{
	SortingList<Term> matches;
	int first = 0;
	int last = 0;
	search(prefix, first, last);
	if (first != -1 && last != -1)
	{
		for(int i = first; i <= last; i++)
		{
			matches.insert(terms[i]);
			matches.merge_sort(Term::compareByWeight);
		}
	}return matches;
}

void Autocomplete::print()
{
	terms.print();
}
